#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.tools
      @file: namespace.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from typing import Any, Dict, Iterator, List, Optional, Tuple

from hspylib.core.preconditions import check_not_none
from hspylib.core.tools.dict_tools import merge

ATTRIBUTE_TYPES = Dict[str, Any] | Tuple[Dict[str, Any]] | List[Dict[str, Any]]


class Namespace:

    @staticmethod
    def of(type_name: str, attributes: ATTRIBUTE_TYPES, final: bool = False) -> 'Namespace':
        check_not_none(attributes)
        self = Namespace(type_name, final)
        self += attributes if isinstance(attributes, dict) else merge(attributes)
        return self

    def __init__(self, type_name: str = None, final: bool = False, **kwargs) -> None:
        self.__name__ = type_name or self.__class__.__name__
        self._final = None
        self._index = 0
        list(map(self.setattr, kwargs.keys(), kwargs.values()))
        self._final = final

    def key(self) -> Tuple[Any]:
        return tuple(self.values)

    def __str__(self) -> str:
        return f"{self.__name__}({', '.join([f'{a}={av}' for a, av in zip(self.attributes, self.values)])})"

    def __repr__(self) -> str:
        return str(self)

    def __hash__(self) -> int:
        return hash(self.key())

    def __eq__(self, other: 'Namespace') -> bool:
        if isinstance(other, self.__class__):
            return self.key() == other.key()
        return NotImplemented

    def __getitem__(self, attribute_name: str) -> Any:
        return getattr(self, attribute_name)

    def __len__(self) -> int:
        return len(self.attributes)

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        self._index = 0
        return self

    def __next__(self) -> Tuple[str, Any]:
        if self._index < len(self):
            item = self.at(self._index)
            self._index += 1
            return item
        raise StopIteration

    def __contains__(self, attribute: str):
        return hasattr(self, attribute)

    def __iadd__(self, attribute: Dict[str, Any] | 'Namespace') -> 'Namespace':
        for a, av in attribute.items():
            self.setattr(a, av)
        return self

    def __add__(self, other: Dict[str, Any] | 'Namespace') -> 'Namespace':
        self.__name__ += '.' + other.__name__
        return self.__iadd__(other)

    def setattr(self, name: str, value: Any) -> 'Namespace':
        if self._final and not hasattr(self, name):
            raise ValueError(f"{self.__name__} Namespace is final")
        if name.startswith(('_', '__')) or name.endswith(('_', '__')):
            raise NameError("Attribute names can't start with '_' or '__'")
        setattr(self, name, value)
        return self

    def hasattr(self, name: str) -> bool:
        return hasattr(self, name)

    def at(self, index: int) -> Optional[Tuple[str, Any]]:
        return (self.attributes[index], self.values[index]) if index < len(self) else None

    def items(self) -> Iterator[Tuple[str, Any]]:
        return iter(self)

    @property
    def attributes(self) -> Tuple[str]:
        return tuple(
            a for a in
            filter(lambda name:
                   self[name] is not None
                   and not name.startswith(('_', '__'))
                   and not name.endswith(('_', '__')),
                   vars(self)))

    @property
    def values(self) -> Tuple[Any]:
        return tuple(
            filter(lambda val:
                   val is not None,
                   [getattr(self, a) for a in self.attributes]))



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.128
#
# Package: main.hspylib.modules
"""Package initialization."""

__all__ = [
  'cache', 
  'cli', 
  'eventbus', 
  'fetch', 
  'json_search', 
  'mock', 
  'qt', 
  'security', 
  'web'
]
__version__ = '0.11.128'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: hspylib.modules.cache
      @file: ttl_keyring_be.py
   @created: Tue, 4 Oct 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import json
from datetime import datetime
from typing import Callable, Optional

import keyring
from keyring.backends.chainer import ChainerBackend
from keyring.errors import PasswordDeleteError, PasswordSetError

from hspylib.core.preconditions import check_not_none
from hspylib.core.zoned_datetime import now_ms
from hspylib.modules.security.security import b64_decode, b64_encode


class TTLKeyringBE(ChainerBackend):

    priority = 1

    def __init__(self, ttl_minutes: int = 15, ttl_seconds: int = 0, cb_expired: Callable = None) -> None:
        super().__init__()
        self._ttl = ttl_minutes, ttl_seconds
        self._cb_expired = cb_expired

    def set_password(self, service: str, username: str, password: str) -> None:
        """Set password for the username of the service."""
        try:
            check_not_none(service, username, password)
            expires_sec = now_ms() + (self._ttl[0] * 60 + self._ttl[1])
            passwd_obj = {'sn': service, 'un': username, 'pw': password, 'ttl': expires_sec}
            b64_pwd = b64_encode(json.dumps(passwd_obj))
            super().set_password(service, username, b64_pwd)
        except PasswordSetError:
            pass  # it does not matter if the password set failed.

    def get_password(self, service: str, username: str) -> Optional[str]:
        """Get password of the username for the service."""
        check_not_none(service, username)
        if b64_pwd := super().get_password(service, username):
            passwd_str = b64_decode(b64_pwd)
            passwd_obj = json.loads(passwd_str)
            dt_object = datetime.fromtimestamp(passwd_obj['ttl'])
            if now_ms() - dt_object.timestamp() > 0:
                keyring.delete_password(service, username)
                if self._cb_expired:
                    self._cb_expired(passwd_obj['pw'])
            else:
                return passwd_obj['pw']

        return None

    def delete_password(self, service: str, username: str) -> None:
        """Delete the password for the username of the service."""
        check_not_none(service, username)
        try:
            super().delete_password(service, username)
        except PasswordDeleteError:
            pass  # it does not matter if the password does not exist.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @package: hspylib.modules.cache
      @file: ttl_cache.py
   @created: Wed, 7 Dec 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import ast
import os
import tempfile
from threading import Lock
from typing import Generic, Optional, TypeVar

import keyring

from hspylib.core.enums.charset import Charset
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.preconditions import check_not_none
from hspylib.core.tools.commons import safe_delete_file
from hspylib.modules.cache.ttl_keyring_be import TTLKeyringBE
from hspylib.modules.security.security import b64_decode, b64_encode

T = TypeVar('T')


class TTLCache(Generic[T], metaclass=Singleton):
    """TODO"""

    CACHE_SERVICE = 'HS-CACHE-SERVICE'

    def __init__(self, ttl_minutes: int = 15, ttl_seconds: int = 0) -> None:
        super().__init__()
        self._lock = Lock()
        keyring.set_keyring(TTLKeyringBE(ttl_minutes, ttl_seconds, safe_delete_file))

    def save(self, key: str, entry: T) -> str:
        """TODO"""
        check_not_none(key, entry)
        with self._lock:
            with tempfile.NamedTemporaryFile(delete=False, mode='w', encoding=Charset.UTF_8.val) as f_temp:
                content = b64_encode(f"{repr(entry)}")
                f_temp.write(content)
                keyring.set_password(self.CACHE_SERVICE, key, f_temp.name)
                return f_temp.name

    def read(self, key: str) -> Optional[T]:
        """TODO"""
        check_not_none(key)
        with self._lock:
            cache_name = keyring.get_password(self.CACHE_SERVICE, key)
            if not cache_name or not os.path.exists(cache_name):
                return None
            try:
                with open(cache_name, encoding=Charset.UTF_8.val) as f_cache:
                    content = b64_decode(f_cache.readline())
                    return ast.literal_eval(content)
            except (ValueError, TypeError, SyntaxError):
                return None

    def delete(self, key: str) -> None:
        """TODO"""
        check_not_none(key)
        with self._lock:
            keyring.delete_password(self.CACHE_SERVICE, key)


if __name__ == '__main__':
    c = TTLCache(ttl_minutes=0, ttl_seconds=10)
    l = [1, 2, 3, 4, 5]
    print(c.save('test', l))
    l2 = c.read('test')
    print(l2)
    m = dict(name= 'test', age= 31)
    print(c.save('test1', m))
    m2 = c.read('test1')
    print(m2)
